{-# LANGUAGE NPlusKPatterns #-}

module Matrix where
import Data.Complex

infixl 7 |.|, |*|, |**|
infixl 6 |@|
infixl 5 |+|, |-|, |++|

type R = Double          --Real number
type C = Complex R       --Complex number

pow x y = exp (y * log x)

data Matrix a = Msca a | Mqua (Matrix a) (Matrix a) (Matrix a) (Matrix a)

matrix :: Matrix a -> [[a]]
matrix (Msca x) = [[x]]
matrix (Mqua a b c d) =
   zipWith (++) (matrix a) (matrix b) ++
   zipWith (++) (matrix c) (matrix d)

makeMatrix :: Num a => Int -> (Int -> Int -> a) -> Matrix a
makeMatrix 0 f = Msca (f 0 0)
makeMatrix (n+1) f | n >= 0 =
   let fa = f
       fb i j = f i (j + 2^n)
       fc i j = f (i + 2^n) j
       fd i j = f (i + 2^n) (j + 2^n)
   in  Mqua (makeMatrix n fa) (makeMatrix n fb)
            (makeMatrix n fc) (makeMatrix n fd)

dim :: Matrix a -> Int
dim (Msca _) = 0
dim (Mqua a b c d) =
   let da = dim a
       db = dim b
       dc = dim c
       dd = dim d
   in  if da == db && db == dc && dc == dd then
          da + 1
       else
          error "dim: invalid matrix"

showMatrix :: Show a => Matrix a -> ShowS
showMatrix m = showm (matrix m)
   where showm [] = id
         showm (l : m) = showl l . showm m
         showl l = ("|" ++) . showr l
         showr [] = (" |\n" ++)
         showr (x : l) = (" " ++) . shows x . showr l

instance Show a => Show (Matrix a) where
   showsPrec _ m = showMatrix m

instance Eq a => Eq (Matrix a) where
   Msca x1 == Msca x2 =
      x1 == x2
   Mqua a1 b1 c1 d1 == Mqua a2 b2 c2 d2 =
      a1 == a2 && b1 == b2 &&
      c1 == c2 && d1 == d2
   _ == _ = error "==: bad pattern matching!"

star :: RealFloat a => Matrix (Complex a) -> Matrix (Complex a)
star (Msca x) = Msca (conjugate x)
star (Mqua a b c d) = Mqua (star a) (star c) (star b) (star d)


(|+|) :: Num a => Matrix a -> Matrix a -> Matrix a
Msca x1 |+| Msca x2 =
   Msca (x1 + x2)
Mqua a1 b1 c1 d1 |+| Mqua a2 b2 c2 d2 =
   Mqua (a1 |+| a2) (b1 |+| b2)
        (c1 |+| c2) (d1 |+| d2)
_ |+| _ = error "|+|: bad pattern matching!"

(|-|) :: Num a => Matrix a -> Matrix a -> Matrix a
Msca x1 |-| Msca x2 =
   Msca (x1 - x2)
Mqua a1 b1 c1 d1 |-| Mqua a2 b2 c2 d2 =
   Mqua (a1 |-| a2) (b1 |-| b2)
        (c1 |-| c2) (d1 |-| d2)
_ |-| _ = error "|-|: bad pattern matching!"

(|.|) :: Num a => a -> Matrix a -> Matrix a
k |.| Msca x =
   Msca (k * x)
k |.| Mqua a b c d =
   Mqua (k |.| a) (k |.| b)
        (k |.| c) (k |.| d)

(|*|) :: Num a => Matrix a -> Matrix a -> Matrix a
Msca x1 |*| Msca x2 =
   Msca (x1 * x2)
Mqua a1 b1 c1 d1 |*| Mqua a2 b2 c2 d2 =
   Mqua (a1 |*| a2 |+| b1 |*| c2) (a1 |*| b2 |+| b1 |*| d2)
        (c1 |*| a2 |+| d1 |*| c2) (c1 |*| b2 |+| d1 |*| d2)
_ |*| _ = error "|*|: bad pattern matching!"

(|@|) :: Num a => Matrix a -> Matrix a -> Matrix a
Msca x |@| m =
   (x |.| m)
Mqua a b c d |@| m =
   Mqua (a |@| m) (b |@| m)
        (c |@| m) (d |@| m)

z  :: Num a => Int -> Matrix a
z n | n == 0    = Msca 0
    | n > 0     = let zp = z (n-1)
                  in  Mqua zp zp zp zp
    | otherwise = error "z: bad pattern matching!"

zero :: (Eq a, Num a) => Matrix a -> Bool
zero (Msca x) = x == 0
zero (Mqua a b c d) = zero a && zero b && zero c && zero d

zeroDiff :: Matrix C -> Matrix C -> Bool
zeroDiff (Msca c1) (Msca c2) = magnitude (c1-c2) < absErr
                               where absErr = 1.0e-10
zeroDiff (Mqua a1 b1 c1 d1) (Mqua a2 b2 c2 d2) = 
             zeroDiff a1 a2 && zeroDiff b1 b2 && zeroDiff c1 c2 && zeroDiff d1 d2
zeroDiff _ _ = error "zeroDiff: bad pattern matching"

i :: Num a => Int -> Matrix a
i n | n == 0    = Msca 1
    | n > 0     = let zp = z (n-1)
                      ip = i (n-1)
                  in Mqua ip zp zp ip
    | otherwise = error "i: bad pattern matching!"

d :: Num a => Int -> Matrix a
d n | n == 0    = Msca 1
    | n > 0     = let zp = z (n-1)
                      dp = d (n-1)
                  in Mqua dp zp zp zp
    | otherwise = error "d: bad pattern matching!"

mnot :: Num a => Matrix a
mnot = Mqua (Msca 0) (Msca 1)
            (Msca 1) (Msca 0)

except :: (Eq a, Num a) => Int -> Matrix a -> Matrix a
except 0 (Mqua a b c d) | a == d && zero b && zero c = a
except (k+1) (Mqua a b c d) | k >= 0 =
   Mqua (except k a) (except k b)
        (except k c) (except k d)
except _ _ = error "except: bad pattern matching!"

cond :: Num a => Int -> Matrix a -> Matrix a -> Matrix a
cond 0 t f =
   let dt = dim t
       df = dim f
       zp = z dt
   in  if dt == df then
          Mqua f zp zp t
       else
          error "cond: invalid dimensions in two matrices"
cond (k+1) (Mqua ta tb tc td) (Mqua fa fb fc fd) | k >= 0 =
   Mqua (cond k ta fa) (cond k tb fb)
        (cond k tc fc) (cond k td fd)
cond _ _ _ = error "cond: bad pattern matching!"

measure :: Num a => Int -> Matrix a -> (Matrix a, Matrix a)
measure 0 (Mqua a _ _ d) =
   let zp = z (dim a)
   in  (Mqua zp zp zp d, Mqua a zp zp zp)
measure (k+1) (Mqua a b c d) | k >= 0 =
   let (ta, fa) = measure k a
       (tb, fb) = measure k b
       (tc, fc) = measure k c
       (td, fd) = measure k d
   in  (Mqua ta tb tc td, Mqua fa fb fc fd)
measure _ _ = error "measure: bad pattern matching!"

measureLast :: Num a => Int -> Matrix a -> Matrix a
measureLast 0 m = m
measureLast (k+1) m | k>=0 = measureLastOne (measureLast k m)  --k+1 iterations

measureLastOne :: Num a => Matrix a -> Matrix a
measureLastOne (Mqua (Msca a) (Msca b) (Msca c) (Msca d)) = Msca (a+d)
measureLastOne (Mqua a b c d) = Mqua (measureLastOne a) (measureLastOne b)
                                     (measureLastOne c) (measureLastOne d)

make a b c d = Mqua (Msca a) (Msca b) (Msca c) (Msca d)

expand :: Num a => Int -> [Int] -> Matrix a -> Matrix a
expand n s t = aux 0 s t
   where aux k s (Mqua a b c d) | k < n && k `elem` s =
            Mqua (aux (k+1) s a) (aux (k+1) s b)
                 (aux (k+1) s c) (aux (k+1) s d)
         aux k s t | k < n && not (k `elem` s) =
            i 1 |@| aux (k+1) s t
         aux k s t | k >= n = t

tr :: Num a => Matrix a -> a
tr (Msca a) = a
tr (Mqua ta tb tc td) = tr ta + tr td

ptr :: Num a => [Int] -> Matrix a -> Matrix a
ptr l m = aux 0 m
  where aux k (Msca a) = Msca a
        aux k (Mqua ta tb tc td) | k `elem` l = Mqua (aux (k+1) ta) (aux (k+1) tb)
                                                     (aux (k+1) tc) (aux (k+1) td)
                                 | otherwise  = aux (k+1) ta |+| aux (k+1) td

--Vectors
data Vector a = Vsca a | Vqua (Vector a) (Vector a)

dimV :: Vector a -> Int
dimV (Vsca _) = 0
dimV (Vqua a b ) =
   let da = dimV a
       db = dimV b
   in  if da == db then
          da + 1
       else
          error "dimV: invalid vector"

(|**|) :: Num a => Matrix a -> Vector a -> Vector a
Msca x1 |**| Vsca v1 =
   Vsca (x1*v1)
Mqua a1 b1 c1 d1 |**| Vqua v1 v2 =
   Vqua (a1 |**| v1 |++| b1 |**| v2) (c1 |**| v1 |++| d1 |**| v2)
_ |**| _ = error "|**|: bad pattern matching!"

(|++|) :: Num a => Vector a -> Vector a -> Vector a
Vsca x1 |++| Vsca x2 =
   Vsca (x1 + x2)
Vqua a1 b1 |++| Vqua a2 b2 =
   Vqua (a1 |++| a2) (b1 |++| b2)
_ |++| _ = error "|++|: bad pattern matching!"

(|@@|) :: Num a => Vector a -> Vector a -> Vector a
Vsca x |@@| m =
   (x |..| m)
Vqua a b |@@| m =
   Vqua (a |@@| m) (b |@@| m)

(|..|) :: Num a => a -> Vector a -> Vector a
k |..| Vsca x =
   Vsca (k * x)
k |..| Vqua a b =
   Vqua (k |..| a) (k |..| b)
   
(|><|) :: Vector C -> Vector C -> Matrix C
Vsca x1 |><| Vsca x2 =
   Msca (x1 * (conjugate x2))
Vqua a1 b1 |><| Vqua a2 b2 =
   Mqua (a1 |><| a2) (a1 |><| b2)
        (b1 |><| a2) (b1 |><| b2)
_ |><| _ = error "|><|: bad pattern matching!"

zeroV :: Num a => Int -> Vector a
zeroV n | n == 0    = Vsca 0
        | n > 0     = let zp = zeroV (n-1)
                      in  Vqua zp zp
        | otherwise = error "zeroV: bad pattern matching!"

oneV :: Num a => Int -> Vector a
oneV n | n==0 = Vsca 1
       | n>0 = let op = oneV (n-1)
                   zp = zeroV (n-1)
               in Vqua op zp
       | otherwise = error "oneV: bad pattern matching!"

showBool :: Show a => String -> Vector a -> String
showBool s (Vsca x) = s ++ ": " ++ show x ++ "\n"
showBool s (Vqua v1 v2) = (showBool (s++" False") v1) ++ (showBool (s++" True ") v2)

instance Show a => Show (Vector a) where
  show v = "\nSTATE VECTOR\n" ++ showBool "" v

density :: Vector C -> Matrix C
density v = v |><| v
