module QCirc where

import Data.Complex
import Matrix

-- Circuits
data Circ = Rot (C,C) (C,C)
          | Wire [Int]
          | Par Circ Circ
          | Seq Circ Circ
          | Cond Circ Circ
          | Unit (Matrix C)
          deriving Eq

arity :: Circ -> Int
arity (Rot x y) | not (orthTest x y) = error $ "Orthogonality: " ++ show (x,y)
                | otherwise          = 1
arity (Wire x) | not (chkPerm x)     = error $ "Wire (permutation error): " ++ show x
               | otherwise           = length x
arity (Par x y)                      = arity x + arity y
arity (Seq x y) | m /= n             = error $ "Seq: arity=" ++ show (m,n)
                | otherwise          = m
  where (m, n) = (arity x, arity y)
arity (Cond x y) | m /= n            = error $ "Cond: arity=" ++ show (m,n)
                 | otherwise         = 1+m
  where (m, n) = (arity x, arity y)
arity (Unit m)                       = dim m

orthTest :: (C,C)->(C,C)->Bool
orthTest (c1,c2) (c3,c4) = conjugate c1 * c3 == - conjugate c2 * c4

chkPerm :: [Int] -> Bool
chkPerm x = and [y `elem` x | y <- [0 .. length x - 1]]

-- Compiling a circuit into a Matrix
comp :: Circ -> Matrix C
comp (Rot (c1,c2) (c3,c4)) = Mqua (Msca c1) (Msca c2) (Msca c3) (Msca c4)
comp (Wire p) =  makeMatrix rcdim f
  where rcdim = length p
        f i j | i == permuteBits p j rcdim = 1
              | otherwise                  = 0
comp (Par c1 c2) = comp c1 |@| comp c2 -- Tensor product
comp (Seq c1 c2) = comp c2 |*| comp c1 -- Inner product
comp (Cond c1 c2) | arity c1 /= arity c2 =
    error ("Cond arity: " ++ show (arity c1, arity c2))
                  | otherwise            = cond 0 (comp c1) (comp c2)
comp (Unit m) = m                      -- returns the unitary matrix

permuteBits :: [Int] -> Int -> Int -> Int
permuteBits [] _ _        = 0
permuteBits (p:ps) x size = s + permuteBits ps newX size
  where radix    = length ps
        isBitSet = x >= 2 ^ radix
        s    = if isBitSet then 2 ^ (size - p - 1) else 0
        newX = if isBitSet then x - 2 ^ radix else x

{-
-- EVALUATING A STATE VECTOR IMMEDIATELY BY THE CIRCUIT
evalCirc :: Circ -> Vec -> Vec
evalCirc (Rot (c1,c2) (c3,c4)) v@(Vec d f) | (d/=1) = error "EvalCirc: Input vector with wrong size (ROT)"
                                           | otherwise = let newf [True] = c1 * (f [True]) + c2 * (f [False])
                                                             newf [False] = c3 * (f [True]) + c4 * (f [False])
                                                         in Vec d newf
evalCirc (Wire p) v@(Vec d f) | (d/= (length p)) = error "EvalCirc: Input vector with wrong size (WIRE)"
                              | otherwise = let newf x = f [x!!q|q<-p]
                                            in Vec d newf
evalCirc (Par c1 c2) v@(Vec d f) | (d/= (arity c1) + (arity c2)) = error "EvalCirc: Input vector with wrong size (PAR)"
                                 | otherwise = v >< ((comp c1) >><< (comp c2))
evalCirc (Seq c1 c2) v@(Vec d f) | ((arity c1)/=(arity c2)) = error "EvalCirc: Circuits' arities don't match (SEQ)"
                                 | (d/=(arity c1)) = error "EvalCirc: Input vector with wrong size (SEQ)"
                                 | otherwise = evalCirc c2 (evalCirc c1 v)
evalCirc (Cond c1 c2) v@(Vec d f) | ((arity c1)/=(arity c2)) = error "EvalCirc: Circuits' arities don't match (COND)"
                                  | (d /= (arity c1) + 1) = error "EvalCirc: Input vector with wrong size (COND)"
                                  | otherwise = let trueFun x = f (True:x)
                                                    falseFun x = f (False:x)
                                                    newf (True:x) = funV (evalCirc c1 (Vec (d-1) trueFun)) x
                                                    newf (False:x) = funV (evalCirc c2 (Vec (d-1) falseFun)) x
                                                in Vec d newf
evalCirc (Unit m@(Mat a b fm)) v@(Vec d f) | (d/=b) = error "EvalCirc: Input vector with wrong size (UNIT)"
                                           | otherwise = v >< m
-}

-- Standard Circuits
idC :: Int -> Circ
idC x = Wire [0..x-1]

-- Not
notC :: Circ
notC = Rot (0,1) (1,0)

-- CNot
cnotC :: Circ
cnotC = Cond notC (idC 1)

-- Hadamard
hadC :: Circ
hadC = Rot (h,h) (h,-h)
  where h = 1 / sqrt 2

-- Toffoli
toffoliC :: Circ
toffoliC = Cond cnotC (idC 2)

-- Mutliple Hadamard
hadnC :: Int -> Circ
hadnC 0 = error "Invalid arity"
hadnC 1 = hadC
hadnC n = Par (hadnC (n-1)) hadC

-- Circuit Optimizer
rCircuit :: Circ -> Circ
rCircuit x | rCirc x == x = x
           | otherwise    = rCircuit (rCirc x)  --Optimization until equal

rCirc :: Circ -> Circ
rCirc (Par x (Wire []))                              = x
rCirc (Par (Wire []) y)                              = y
rCirc (Par (Wire x) (Wire y))                        = Wire( x ++ [ys + length x| ys<-y])
rCirc (Par x y)                                      = Par (rCirc x) (rCirc y)
rCirc (Seq (Wire x) (Wire y))                        = Wire (seqPerm x y)
rCirc (Seq c (Wire x)) | x == [0..length x - 1]      = rCirc c
                       | otherwise                   = Seq (rCirc c) (Wire x)
rCirc (Seq (Wire x) c) | x == [0..length x - 1]      = rCirc c
                       | otherwise                   = Seq (Wire x) (rCirc c)
rCirc (Seq (Par x y) (Par a b)) | arity x == arity a = rCirc (Par (Seq x a) (Seq y b))
rCirc (Seq x y)                                      = Seq (rCirc x) (rCirc y)
rCirc (Cond x y) | x==y                              = Par (idC 1) (rCirc x)
                 | otherwise                         = Cond (rCirc x) (rCirc y)
rCirc c                                              = c -- nothing to do!

seqPerm :: [Int] -> [Int] -> [Int]
seqPerm x y | lx /= ly     = error "Arity"
            | x == [0..lx] = y     -- x is id
            | y == [0..ly] = x
            | otherwise    = sortp x y
  where lx = length x - 1
        ly = length y - 1
        sortp _  []      = error "sortp"
        sortp xs [yh]    = [xs!!yh]
        sortp xs (yh:ys) = (xs!!yh) : sortp xs ys

cross :: Int -> Int -> Circ
cross a b = Wire $ [ x+b | x <- [0..a-1] ] ++ [0..b-1]

empty :: Circ
empty = Wire []

isEmpty :: Circ -> Bool
isEmpty (Wire [])    = True
isEmpty (Par c1 c2)  = isEmpty c1 && isEmpty c2
isEmpty (Seq c1 c2)  = isEmpty c1 && isEmpty c2
isEmpty (Cond c1 c2) = isEmpty c1 && isEmpty c2
isEmpty _            = False

parc :: [Circ] -> Circ
parc []                   = empty
parc (c : cs) | isEmpty c = parc cs
              | otherwise = Par c (parc cs)

seqc :: [Circ] -> Circ
seqc [c]      = c
seqc (c : cs) = Seq c (seqc cs)

-- Printing
instance Show Circ where
  show (Par x y)               = "Par " ++ showp x y
  show (Seq x y)               = "Seq " ++ showp x y
  show (Wire x)                = "Wire " ++ show x
  show (Cond x y) | x == notC && (y == idC 1 || y == Rot (1,0) (0,1))
                               = "Controlled-Not"
                  | otherwise  = "Cond " ++ showp x y
  show r@(Rot x y) | r == notC = "Not"
                   | r == hadC = "Hadamard"
                   | otherwise = "Rot " ++ show x ++ " " ++ show y
  show (Unit m)                = show (dim m) ++ "-qubit " ++ "unitary"

showp :: Show a => a -> a -> String
showp x y = "(" ++ show x ++ ") (" ++ show y ++ ")"
