module QFqc where

import Matrix
import QCirc

import Data.List  (elemIndex, sort)
import Data.Maybe (fromJust)

data Fqc = Fqc {a :: Int, b :: Int, h :: Int, g :: Int, revCirc::Circ}

validFqc :: Fqc -> Bool
validFqc f = let ah = a f + h f
                 bg = b f + g f
                 res | ah /= bg                = error "Invalid FQC: Input - output arity mismatch"
                     | ah /= arity (revCirc f) = error "Invalid FQC: Circuit arity mismatch"
                     | otherwise               = True
             in res

seqFqc :: Fqc -> Fqc -> Fqc
seqFqc (Fqc a1 b1 h1 g1 c1) (Fqc a2 b2 h2 g2 c2)
    | b1 /= a2 = error "Not valid Seq: Outputs - Inputs mismatch"
    | b1 == 0  = error "Not valid Seq: No outputs of the first circuit"
    | otherwise = Fqc a1 b2 (h1+h2) (g1+g2) resCirc
  where r1 = parc [c1, idC h2]
        r2 = parc [idC b1, cross g1 h2]
        r3 = parc [c2, idC g1]
        r4 = parc [idC b2, cross g2 g1]
        resCirc = rCircuit $ seqc [r1, r2, r3, r4]

parFqc :: Fqc -> Fqc -> Fqc
parFqc (Fqc a1 b1 h1 g1 c1) (Fqc a2 b2 h2 g2 c2) = Fqc (a1+a2) (b1+b2) (h1+h2) (g1+g2) resCirc
    where r1 = parc [idC a1, cross a2 h1, idC h2]
          r2 = parc [c1, c2]
          r3 = parc [idC b1, cross g1 b2, idC g2]
          resCirc = rCircuit $ seqc [r1, r2, r3]

--Evaluation
fqc2Mat :: Fqc -> Matrix C
fqc2Mat (Fqc a b h g c) = comp c

eval :: Fqc -> Vector C -> Vector C
eval r@(Fqc a b h g c) v | (not . validFqc) r = error "Eval: invalid FQC"
                         | a /= dimV v        = error "Eval: Wrong sized input vector"
                         | otherwise          = comp c |**| v
  where dim = dimV v
        inp | dim>0 && h>0 = zeroV h |@@| v   --Appending zeros in the end through tensor product
            | dim==0       = oneV h
            | h==0         = v

evalNoInp :: Fqc -> Vector C
evalNoInp r@(Fqc a b h g c) | (not . validFqc) r = error "EvalNoInp: invalid FQC"
                            | a /= 0             = error "EvalNoInp: input needed by the circuit"
                            | otherwise          = comp c |**| oneV h

fqc2Den :: Fqc -> Matrix C
fqc2Den r@(Fqc a b h g c) = measureLast g $ (density . evalNoInp) r

--N QML CIRCUITS
varFqc :: Int -> Fqc
varFqc n = Fqc n n 0 0 (idC n)  --n are all the qubits

supFqc :: Int -> C -> C -> Fqc
supFqc n c1 c2 | orthTest (c1,c2) (c2,-c1) = Fqc n (n+1) 1 0 $ parc [idC n, Rot (c1,c2) (c2,-c1)]
               | otherwise                 = error "SupFqc: Orthogonality"

nseqFqc :: Fqc -> Fqc -> Fqc
nseqFqc e1@(Fqc a1 b1 h1 g1 c1) e2@(Fqc a2 b2 h2 g2 c2)
    | b1 /= a2 = error "nseqFqc: second expression needs the same inputs"
    | b1 == a2 = seqFqc e1 e2

ifFqc :: Int -> Fqc -> Fqc -> Fqc -> Int -> Fqc
ifFqc n e@(Fqc a b h g c) e1@(Fqc a1 b1 h1 g1 c1) e2@(Fqc a2 b2 h2 g2 c2) k
    | g1 /= 0   = error "ifFqc: First sub circuit is not strict"
    | g2 /= 0   = error "ifFqc: Second sub circuit is not strict"
    | otherwise = Fqc a (b + max h1 h2) (h + max h1 h2) g resCirc
  where h12 = max h1 h2
        r1  = parc [c, idC h12]
        r2  = Wire $ [1..k] ++ [0] ++ [k+1..b-1] ++
                     [x+b+h12 | x <- [0..g-1]] ++ [b..b+h12-1]
        c1' = ext (b-1+h12) $ exceptC k c1
        c2' = ext (b-1+h12) $ exceptC k c2
        r3  = parc [Cond c1' c2', idC g]
        r4  = Wire $ [k] ++ [0..k-1] ++ [k+1..b + h12 + g-1]
        resCirc = rCircuit $ seqc [r1, r2, r3, r4]

ifmFqc :: Int -> Fqc -> Fqc -> Fqc -> Int -> Fqc
ifmFqc n e@(Fqc a b h g c) e1@(Fqc a1 b1 h1 g1 c1) e2@(Fqc a2 b2 h2 g2 c2) k =
    Fqc a (b + h12 - g1 - g2) (h + h12 + 1) (g+g1+g2+1) resCirc
  where h12 = max h1 h2
        r1  = parc [c, idC (h12 + 1)]
        r2  = Wire $ [2..k+1] ++ [0] ++ [k+2..b] ++
                     [x+b+1+h12 | x <- [0..g-1]] ++ [x+b+1 | x <- [0..h12-1]] ++ [1]
        r3  = parc [cnotC, idC (b-1+g+h12)]
        r4  = Wire $ [0] ++ [k+1] ++ [1..k] ++ [k+2..b+g+h12]
        c1' = ext (b+h12) c1
        c2' = ext (b+h12) c2
        r5  = parc [Cond c1' c2', idC g]
        r6  = Wire $ [b+h12] ++ [0..b+h12-1] ++ [x+b+h12+1 | x <- [0..g-1]]
        resCirc = rCircuit $ seqc [r1, r2, r3, r4, r5, r6]

transFqc :: Int -> Fqc -> Matrix C -> [Int] -> Fqc
transFqc n (Fqc a b h g r1) m t
    | length t == 1 && dim m == 1 = let Mqua (Msca c1) (Msca c2) (Msca c3) (Msca c4) = m
                                        r2 = Rot (c1,c2) (c3,c4)
                                        k  = head t
                                        r3 = parc [idC k, r2, idC (n-k-1), idC g]  --applying the transformation to the correct qubit
                                        resCirc = rCircuit $ seqc [r1, r3]
                                    in (Fqc a b h g resCirc)
    | length t /= 1 && dim m /= 1 = let perm = fixPerm n t
                                        r2 = Wire perm
                                        r3 = parc [Unit m, idC (n-length t)]
                                        r4 = Wire (reversePerm perm)
                                        resCirc = rCircuit $ seqc [r1, r2, r3, r4]
                                    in (Fqc a b h g resCirc)

--Auxiliary function
exceptC :: Int -> Circ -> Circ
exceptC k (Par c1 c2) | k == 0 && c1==idC 1 = c2
                      | k == 0              = error "exceptC: cannot extract Wire [0]"
                      | k <  arity c1       = parc [exceptC k c1, c2]
                      | k >= arity c1       = parc [c1, exceptC (k - arity c1) c2]
exceptC k (Wire p) = Wire $ map (\x -> if x >= t then x-1 else x) res --subtracts 1 from the wires going above k
  where (i,f) = splitAt k p         --cuts the permutations into two pieces
        res   = i ++ tail f         --removes k element
        t     = p!!k                --target wire of k wire
exceptC 0 (Cond c1 c2)      = error "exceptC: qubit to except is used in cond"
exceptC k (Cond c1 c2)      = Cond (exceptC (k-1) c1) (exceptC (k-1) c2)
exceptC k (Seq (Wire p) c2) = rCircuit $ Seq (exceptC k (Wire p)) (exceptC (p!!k) c2)
exceptC k (Seq c1 c2)       = rCircuit $ Seq (exceptC k c1) (exceptC k c2)
exceptC k c                 = error "exceptC: circuit must not use the k qubit"

ext :: Int -> Circ -> Circ
ext k c =
  let a = arity c
  in  if a == k then c 
      else if a < k
           then Par c (idC (k - a))
           else error "ext: cannot extend a bigger circuit"

fixPerm :: Int -> [Int] -> [Int]
fixPerm n t = map (\x -> fromJust $ lookup x pAll) [0..n-1]
  where p1   = zip t [0..length t-1]             -- List with indexes
        p2   = makeAuxList 0 (length t) n (sort t) -- Creating list for the rest qubits
        pAll = p1 ++ p2                          -- All the target qubits

makeAuxList :: Int -> Int -> Int -> [Int] -> [(Int,Int)]
makeAuxList j offs n (i:ts) | j == n = []
                            | j <  i = (j,j+offs) : makeAuxList (j+1) offs n (i:ts)
                            | j == i = makeAuxList (j+1) (offs-1) n ts
makeAuxList j offs n [] | j == n    = []
                        | otherwise = (j,j+offs) : makeAuxList (j+1) offs n []

reversePerm :: [Int] -> [Int]
reversePerm t = map (\x -> fromJust $ elemIndex x t) [0..length t - 1]

-- Printing
instance Show Fqc where
  show (Fqc a b h g c) = show . concat $ ["FQC circuit \n", show a, " inputs, "
                                         , show b, " outputs, ", show h
                                         , " heap inputs, ", show g, " garbage outputs\n"
                                         , show c]